# WARNING: auto-generated file. Do NOT edit.
# If this file needs to be edited, change `codegen_kewords_in_args.py` and rerun.
KEYWORD_NAME_TO_KEYWORD_INDEX = {
    "runkeywords": 1,  # run_keywords
    "repeatkeyword": 2,  # repeat_keyword
    "runkeyword": 1,  # run_keyword
    "runkeywordandcontinueonfailure": 1,  # run_keyword_and_continue_on_failure
    "runkeywordandexpecterror": 2,  # run_keyword_and_expect_error
    "runkeywordandignoreerror": 1,  # run_keyword_and_ignore_error
    "runkeywordandreturn": 1,  # run_keyword_and_return
    "runkeywordandreturnif": 2,  # run_keyword_and_return_if
    "runkeywordandreturnstatus": 1,  # run_keyword_and_return_status
    "runkeywordandwarnonfailure": 1,  # run_keyword_and_warn_on_failure
    "runkeywordif": 2,  # run_keyword_if
    "runkeywordifalltestspassed": 1,  # run_keyword_if_all_tests_passed
    "runkeywordifanytestsfailed": 1,  # run_keyword_if_any_tests_failed
    "runkeywordiftestfailed": 1,  # run_keyword_if_test_failed
    "runkeywordiftestpassed": 1,  # run_keyword_if_test_passed
    "runkeywordiftimeoutoccurred": 1,  # run_keyword_if_timeout_occurred
    "runkeywordunless": 2,  # run_keyword_unless
    "waituntilkeywordsucceeds": 3,  # wait_until_keyword_succeeds
}
KEYWORD_NAME_TO_CONDITION_INDEX = {
    "continueforloopif": 1,  # continue_for_loop_if
    "exitforloopif": 1,  # exit_for_loop_if
    "passexecutionif": 1,  # pass_execution_if
    "returnfromkeywordif": 1,  # return_from_keyword_if
    "runkeywordandreturnif": 1,  # run_keyword_and_return_if
    "runkeywordif": 1,  # run_keyword_if
    "runkeywordunless": 1,  # run_keyword_unless
    "setvariableif": 1,  # set_variable_if
    "shouldbetrue": 1,  # should_be_true
    "shouldnotbetrue": 1,  # should_not_be_true
    "skipif": 1,  # skip_if
}
